/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2007 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This program is free software: you can redistribute it and/or modify |
  | it under the terms of the GNU General Public License as published by |
  | the Free Software Foundation, either version 3 of the License, or    |
  | (at your option) any later version.                                  |
  |                                                                      |
  | This program is distributed in the hope that it will be useful,      |
  | but WITHOUT ANY WARRANTY; without even the implied warranty of       |
  | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        |
  | GNU General Public License for more details.                         |
  |                                                                      |
  | You should have received a copy of the GNU General Public License    |
  | along with this program.  If not, see <http://www.gnu.org/licenses/>.|
  +----------------------------------------------------------------------+
  | Author: Valeriu Palos <valeriu@palos.ro>                             |
  +----------------------------------------------------------------------+
*/

/* $Id: header,v 1.16.2.1.2.1 2007/01/01 19:32:09 iliaa Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "php_psstream.h"
#include "moov.h"
#include "php_main.h"
#include "SAPI.h"

static int le_psstream;

PHP_INI_BEGIN()
PHP_INI_ENTRY("psstream.bandwidth_limit", "0", PHP_INI_ALL, NULL)
PHP_INI_ENTRY("psstream.bandwidth_chunk_size", "204800", PHP_INI_ALL, NULL)
PHP_INI_ENTRY("psstream.bandwidth_chunk_interval", "0.2", PHP_INI_ALL, NULL)
PHP_INI_END()

zend_function_entry psstream_functions[] = {
    PHP_FE(psstream_mp4,NULL)
    PHP_FE(psstream_flv,NULL)
    {NULL, NULL, NULL}
};

zend_module_entry psstream_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
    STANDARD_MODULE_HEADER,
#endif
    "psstream",
    psstream_functions,
    PHP_MINIT(psstream),
    PHP_MSHUTDOWN(psstream),
    PHP_RINIT(psstream),
    PHP_RSHUTDOWN(psstream),
    PHP_MINFO(psstream),
#if ZEND_MODULE_API_NO >= 20010901
    "0.1",
#endif
    STANDARD_MODULE_PROPERTIES
};

#ifdef COMPILE_DL_PSSTREAM
ZEND_GET_MODULE(psstream)
#endif


// Startup/Shutdown functions

PHP_MINIT_FUNCTION(psstream)
{
    REGISTER_INI_ENTRIES();
    return SUCCESS;
}

PHP_MSHUTDOWN_FUNCTION(psstream)
{
    UNREGISTER_INI_ENTRIES();
    return SUCCESS;
}

PHP_RINIT_FUNCTION(psstream)
{
    return SUCCESS;
}

PHP_RSHUTDOWN_FUNCTION(psstream)
{
    return SUCCESS;
}

PHP_MINFO_FUNCTION(psstream)
{
    php_info_print_table_start();
    php_info_print_table_header(2, "psstream support", "enabled");
    php_info_print_table_row(2, "psstream version", "1.0");
    php_info_print_table_end();

    DISPLAY_INI_ENTRIES();
}


// Useful macros

#define INIT_HEADERS sapi_header_line psstream_sapi_header_line = {0}

#define ADD_HEADER(h) \
    psstream_sapi_header_line.line = h; \
    psstream_sapi_header_line.line_len = strlen(psstream_sapi_header_line.line); \
    sapi_header_op(SAPI_HEADER_ADD, &psstream_sapi_header_line);

#define REPLACE_HEADER(h) \
    psstream_sapi_header_line.line = h; \
    psstream_sapi_header_line.line_len = strlen(psstream_sapi_header_line.line); \
    sapi_header_op(SAPI_HEADER_REPLACE, &psstream_sapi_header_line);

#define SEND_HEADERS sapi_send_headers();

#define STREAMING_ERROR(t) php_error(E_WARNING, "STREAMING ERROR: %s", t)


// Helper functions

#define ATOM_PREAMBLE_SIZE 8

void write_char(unsigned char* outbuffer, int value) {
    outbuffer[0] = (unsigned char)(value);
}

void write_int32(unsigned char* outbuffer, long value) {
    outbuffer[0] = (unsigned char)((value >> 24) & 0xff);
    outbuffer[1] = (unsigned char)((value >> 16) & 0xff);
    outbuffer[2] = (unsigned char)((value >> 8) & 0xff);
    outbuffer[3] = (unsigned char)((value >> 0) & 0xff);
}

struct atom_t {
    unsigned char type_[4];
    uint64_t size_;
    uint64_t start_;
    uint64_t end_;
};

unsigned int atom_header_size(unsigned char* atom_bytes) {
    return (atom_bytes[0] << 24) +
           (atom_bytes[1] << 16) +
           (atom_bytes[2] << 8) +
           (atom_bytes[3]);
}

int atom_read_header(FILE* infile, struct atom_t* atom) {
    unsigned char atom_bytes[ATOM_PREAMBLE_SIZE];

    atom->start_ = ftell(infile);

    if (!fread(atom_bytes, ATOM_PREAMBLE_SIZE, 1, infile)) {
        return 0;
    }

    memcpy(&atom->type_[0], &atom_bytes[4], 4);
    atom->size_ = atom_header_size(atom_bytes);
    atom->end_ = atom->start_ + atom->size_;

    return 1;
}

void atom_write_header(unsigned char* outbuffer, struct atom_t* atom) {
    int i;
    write_int32(outbuffer, atom->size_);

    for(i = 0; i != 4; ++i) {
        write_char(outbuffer + 4 + i, atom->type_[i]);
    }
}

int atom_is(struct atom_t const* atom, const char* type) {
    return (atom->type_[0] == type[0] &&
            atom->type_[1] == type[1] &&
            atom->type_[2] == type[2] &&
            atom->type_[3] == type[3]);
}

void atom_skip(FILE* infile, struct atom_t const* atom) {
    fseek(infile, atom->end_, SEEK_SET);
}

void atom_print(struct atom_t const* atom) {
    printf("Atom(%c%c%c%c,%lld)\n", atom->type_[0], atom->type_[1],
        atom->type_[2], atom->type_[3], atom->size_);
}

double precise_time() {
    struct timeval tp = {0};
    struct timezone tz = {0};

    if (gettimeofday(&tp, &tz))
            return 0;

    return (double)(tp.tv_sec + tp.tv_usec) / 1000000.00;
}


// Main functions

/* MP4 (H264) pseudo-streaming.
 * This function does nothing to check if the requested file is really a valid MP4/H264 file.
 */
PHP_FUNCTION(psstream_mp4)
{
    char *path;
    unsigned long path_size;
    double t_start = 0.0;
    double t_end = 0.0;

    if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|d", &path, &path_size, &t_start) == FAILURE)
        WRONG_PARAM_COUNT;

    if (t_start < 0) {
        php_error(E_WARNING, "Invalid start offset!");
        RETURN_FALSE;
    }

    FILE* infile;
    struct atom_t ftyp_atom;
    struct atom_t moov_atom;
    struct atom_t mdat_atom;
    unsigned char* moov_data = 0;
    unsigned char* ftyp_data = 0;
    struct stat filestat;

    if(VCWD_STAT(path, &filestat) || !(infile = VCWD_FOPEN(path, "rb"))) {
        php_error(E_WARNING, "Could not open file... [%s]", path);
        RETURN_FALSE;
    }
    unsigned long filesize = filestat.st_size;

    if (SG(headers_sent)) {
        php_error(E_WARNING, "Can not start streaming: headers were already sent!");
        RETURN_FALSE;
    }

    // Send H264 structure
    struct atom_t leaf_atom;

    while(ftell(infile) < filesize) {
        if(!atom_read_header(infile, &leaf_atom))
            break;

        atom_print(&leaf_atom);

        if(atom_is(&leaf_atom, "ftyp")) {
            ftyp_atom = leaf_atom;
            ftyp_data = malloc(ftyp_atom.size_);
            fseek(infile, ftyp_atom.start_, SEEK_SET);
            if (!fread(ftyp_data, ftyp_atom.size_, 1, infile)) {
                STREAMING_ERROR("file read error");
                RETURN_FALSE;
            }
        }
        else if(atom_is(&leaf_atom, "moov")) {
            moov_atom = leaf_atom;
            moov_data = malloc(moov_atom.size_);
            fseek(infile, moov_atom.start_, SEEK_SET);
            if (!fread(moov_data, moov_atom.size_, 1, infile)) {
                STREAMING_ERROR("file read error");
                RETURN_FALSE;
            }
        }
        else if(atom_is(&leaf_atom, "mdat")) {
            mdat_atom = leaf_atom;
        }
        atom_skip(infile, &leaf_atom);
    }
    fseek(infile, 0, SEEK_SET);

    if(!moov_data) {
        STREAMING_ERROR("null/empty moov_data");
        RETURN_FALSE;
    }

    unsigned int mdat_start = (ftyp_data ? ftyp_atom.size_ : 0) + moov_atom.size_;

    if(!moov_seek(moov_data,
            &moov_atom.size_,
            t_start, t_end,
            &mdat_atom.start_, &mdat_atom.size_,
            mdat_start - mdat_atom.start_)) {
        STREAMING_ERROR("moov_seek failed");
        RETURN_FALSE;
    }

    // Compute start/end file offsets
    unsigned long start = mdat_atom.start_ + ATOM_PREAMBLE_SIZE;
    unsigned long end = start + mdat_atom.size_ - ATOM_PREAMBLE_SIZE;

    // Send headers
    char last_modified[200];
    time_t t = time(NULL);
    struct tm *tmp;
    if (tmp = localtime(&t)) {
        strftime(last_modified, sizeof(last_modified), "Last-Modified: %a, %d %B %y %H:%M:%S GMT", tmp);
    }

    unsigned long delta = end - start;
    delta += ftyp_data ? ftyp_atom.size_ : 0;
    delta += moov_atom.size_;
    delta += ATOM_PREAMBLE_SIZE;

    char content_length[100];
    sprintf(content_length, "Content-Length: %lu", delta);

    INIT_HEADERS;
    ADD_HEADER("Content-Type: video/mp4");
    ADD_HEADER("Expires: Thu, 19 Nov 1981 08:52:00 GMT");
    ADD_HEADER(last_modified);
    ADD_HEADER("Cache-Control: no-store, no-cache, must-revalidate, post-check=0, pre-check=0");
    ADD_HEADER("Pragma: no-cache");
    ADD_HEADER(content_length);
    SEND_HEADERS;

    // Send meta atoms
    if(ftyp_data) {
        PHPWRITE(ftyp_data, ftyp_atom.size_);
        free(ftyp_data);
    }

    PHPWRITE(moov_data, moov_atom.size_);
    free(moov_data);

    unsigned char mdat_bytes[ATOM_PREAMBLE_SIZE];
    atom_write_header(mdat_bytes, &mdat_atom);
    PHPWRITE(mdat_bytes, ATOM_PREAMBLE_SIZE);

    // Configure options
    int bandwidth_limit = INI_BOOL("psstream.bandwidth_limit");
    unsigned long bandwidth_chunk_size = INI_INT("psstream.bandwidth_chunk_size");
    double bandwidth_chunk_interval = INI_FLT("psstream.bandwidth_chunk_interval");

    if (bandwidth_chunk_size < 1024) bandwidth_chunk_size = 1024;
    else if (bandwidth_chunk_size > 1048576) bandwidth_chunk_size = 1048576;

    if (bandwidth_chunk_interval < 0.1) bandwidth_chunk_interval = 0.1;
    else if (bandwidth_chunk_interval > 2) bandwidth_chunk_interval = 2;

    unsigned long chunk_size = bandwidth_limit ? bandwidth_chunk_size : 204800;

    // Set some ini settings
    zend_alter_ini_entry("session.cache_limiter", sizeof("session.cache_limiter"),
        "nocache", sizeof("nocache"), PHP_INI_USER, PHP_INI_STAGE_RUNTIME);

    // Stream data
    double t_delta = 0.0;
    unsigned long result;

    unsigned char *buffer = (unsigned char*)malloc(chunk_size);
    if (!buffer) {
        RETURN_FALSE;
    }

    fseek(infile, start, SEEK_SET);
    while(start < end) {

        if (end - start < chunk_size) {
            chunk_size = end - start;
        }

        t_start = precise_time();

        result = fread(buffer, 1, chunk_size, infile);

        if (result != chunk_size) {
            free(buffer);
            RETURN_FALSE;
        }

        PHPWRITE(buffer, chunk_size);
        start += chunk_size;

        if(bandwidth_limit) {
            t_end = precise_time();
            t_delta = t_end - t_start;

            if(t_delta < bandwidth_chunk_interval) {
                usleep(bandwidth_chunk_interval * 1000000 - t_delta * 1000000);
            }
        }
    }

    // Close file
    fclose(infile);
    free(buffer);

    RETURN_TRUE;
}

/* Classic FLV pseudo-streaming.
 * This function does nothing to check if the requested file is really a valid FLV file.
 */
PHP_FUNCTION(psstream_flv)
{
    #define FLV_SIGNATURE "FLV\x1\x1\0\0\0\x9\0\0\0\x9"

    char *path;
    unsigned long path_size;
    unsigned long start = 0;

    if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|l", &path, &path_size, &start) == FAILURE)
        WRONG_PARAM_COUNT;

    FILE* infile;
    struct stat filestat;

    if(VCWD_STAT(path, &filestat) || !(infile = VCWD_FOPEN(path, "rb"))) {
        php_error(E_WARNING, "Could not open file... [%s]", path);
        RETURN_FALSE;
    }
    unsigned long end = filestat.st_size;

    if (start < 0 || start >= end) {
        php_error(E_WARNING, "Invalid start offset!");
        RETURN_FALSE;
    }

    if (SG(headers_sent)) {
        php_error(E_WARNING, "Can not start streaming: headers were already sent!");
        RETURN_FALSE;
    }

    // Send headers
    char last_modified[200];
    time_t t = time(NULL);
    struct tm *tmp;
    if (tmp = localtime(&t)) {
        strftime(last_modified, sizeof(last_modified), "Last-Modified: %a, %d %B %y %H:%M:%S GMT", tmp);
    }

    char content_length[100];
    sprintf(content_length, "Content-Length: %lu", end - start + sizeof(FLV_SIGNATURE) - 1);

    INIT_HEADERS;
    ADD_HEADER("Content-Type: video/x-flv");
    ADD_HEADER("Expires: Thu, 19 Nov 1981 08:52:00 GMT");
    ADD_HEADER(last_modified);
    ADD_HEADER("Cache-Control: no-store, no-cache, must-revalidate, post-check=0, pre-check=0");
    ADD_HEADER("Pragma: no-cache");
    ADD_HEADER(content_length);
    SEND_HEADERS;

    // Send FLV signature
    if (start > 0) {
        PHPWRITE(FLV_SIGNATURE, sizeof(FLV_SIGNATURE) - 1);
    }

    // Configure options
    int bandwidth_limit = INI_BOOL("psstream.bandwidth_limit");
    unsigned long bandwidth_chunk_size = INI_INT("psstream.bandwidth_chunk_size");
    double bandwidth_chunk_interval = INI_FLT("psstream.bandwidth_chunk_interval");

    if (bandwidth_chunk_size < 1024) bandwidth_chunk_size = 1024;
    else if (bandwidth_chunk_size > 1048576) bandwidth_chunk_size = 1048576;

    if (bandwidth_chunk_interval < 0.1) bandwidth_chunk_interval = 0.1;
    else if (bandwidth_chunk_interval > 2) bandwidth_chunk_interval = 2;

    unsigned long chunk_size = bandwidth_limit ? bandwidth_chunk_size : 204800;

    // Set some ini settings
    zend_alter_ini_entry("session.cache_limiter", sizeof("session.cache_limiter"),
        "nocache", sizeof("nocache"), PHP_INI_USER, PHP_INI_STAGE_RUNTIME);

    // Stream data
    double t_start = 0.0;
    double t_end = 0.0;
    double t_delta = 0.0;
    unsigned long result;

    unsigned char *buffer = (unsigned char*)malloc(chunk_size);
    if (!buffer) {
        RETURN_FALSE;
    }

    fseek(infile, start, SEEK_SET);
    while(start < end) {

        if (end - start < chunk_size) {
            chunk_size = end - start;
        }

        t_start = precise_time();

        result = fread(buffer, 1, chunk_size, infile);

        if (result != chunk_size) {
            free(buffer);
            RETURN_FALSE;
        }

        PHPWRITE(buffer, chunk_size);
        start += chunk_size;

        if(bandwidth_limit) {
            t_end = precise_time();
            t_delta = t_end - t_start;

            if(t_delta < bandwidth_chunk_interval) {
                usleep(bandwidth_chunk_interval * 1000000 - t_delta * 1000000);
            }
        }
    }

    // Close file
    fclose(infile);
    free(buffer);

    RETURN_TRUE;
}